local famBags = LibStub('AceAddon-3.0'):GetAddon('famBags')
famBags.Classification = {}

-- :::::::::::: LOCALS ::::::::::::: -------------------------------------------------------------------------

local wrong_criterias = {}

-- :::::::::::: famBags.Classification ::::::::::::: -------------------------------------------------------------------------
function famBags.Classification:GetButtonCategory(info)
   local db = famBags.db.profile.categories
   local cat, result, val
   if info.id then
      for _, catname in pairs(famBags.SortedCategories) do
         cat = db[catname]
         if cat.Items and cat.Items[info.id] then 
            return catname
         end
      end
   end
   for _, catname in pairs(famBags.SortedCategories) do
      cat = db[catname]
      result = info.id and self:Types(cat, info)
      if cat.Criterias then
         for i, crit in ipairs(cat.Criterias) do
--            dump('crit', crit)
--            dump('info', info)
            if not self[crit.name] then
               if not tContains(wrong_criterias, crit.name) then
                  print(string.format('|cffff0000famBags error: |r \'%s\' criteria not found', crit.name))
                  table.insert(wrong_criterias, crit.name)
               end
               return nil
            end
            val = self[crit.name](crit, info)
            if result == nil then 
               result = val
            elseif crit.operator == 'AND' then
               result = result and val
            elseif crit.operator == 'AND NOT' then
               result = result and not val
            elseif crit.operator == 'OR' then
               result = result or val
            elseif crit.operator == 'OR NOT' then
               result = result or not val
            end
         end
      end
      if result then return catname end
   end
   return nil
end

function famBags.Classification:Types(cat, info)
   if cat.Types and next(cat.Types) then
      if not info.type then return false end
      return (rawget(cat.Types, info.type) and 
             (rawget(cat.Types[info.type], famBags.L['.: All :.']) or 
              rawget(cat.Types[info.type], info.subType))) or false
   else
      return nil
   end
   return false
end

famBags.Classification[famBags.L['Others']] = function(crit, info)
   return true
end

famBags.Classification[famBags.L['Empty']] = function(crit, info)
   return (not info.id) or false
end

famBags.Classification[famBags.L['Quest']] = function(crit, info)
   return info.isQuestItem or false
end
              
famBags.Classification[famBags.L['Usable']] = function(crit, info)
   return info.spell or false
end

famBags.Classification[famBags.L['Key']] = function(crit, info)
   return info.bag == KEYRING_CONTAINER
end

famBags.Classification[famBags.L['New']] = function(crit, info)
   return info.new or false
end

famBags.Classification[famBags.L['Incremented']] = function(crit, info)
   return info.added or false
end

famBags.Classification[famBags.L['Quality']] = function(crit, info)
   if not info.quality then return false end
   local quality = (info.quality >= 0) and info.quality..'.'.._G[string.format('ITEM_QUALITY%s_DESC', info.quality)] or famBags.L['unknown']
   return (rawget(crit.criterias, famBags.L['.: All :.']) or rawget(crit.criterias, quality)) or false
end

famBags.Classification[famBags.L['Equipment Set']] = function(crit, info)
   for set, items in pairs(famBags.EquipmentSets) do
      if items and (rawget(crit.criterias, famBags.L['.: All :.']) or rawget(crit.criterias, set)) and rawget(items, info.id) then
         return true
      end      
   end 
   return false
end

famBags.Classification[famBags.L['Outfitter Set']] = function(crit, info)
   if not famBags.OUTFITTER then return false end
   for set, items in pairs(famBags.OutfitterSets) do
      if items and (rawget(crit.criterias, famBags.L['.: All :.']) or rawget(crit.criterias, set)) and rawget(items, info.id) then
         return true
      end
   end
   return false
end

famBags.Classification[famBags.L['Equipment Slot']] = function(crit, info)
   return (info.equipLoc and info.equipLoc ~= '' and 
          (rawget(crit.criterias, famBags.L['.: All :.']) or rawget(crit.criterias, _G[info.equipLoc]))) or false
end

famBags.Classification[famBags.L['Bound Status']] = function(crit, info)
   return (info.bound and info.bound ~= '' and 
          (rawget(crit.criterias, famBags.L['.: All :.']) or rawget(crit.criterias, info.bound))) or false
end

famBags.Classification[famBags.L['Name']] = function(crit, info)
   local operator = next(crit.criterias)
   if not (info.name and crit.value and operator) then return false end
   if operator == famBags.L['Equal'] then
      return info.name == crit.value
   elseif operator == famBags.L['Contains'] then
      return strfind(info.name, crit.value) or false
   elseif operator == famBags.L['Starts with'] then
      return strfind(info.name, '^'..crit.value) or false
   elseif operator == famBags.L['Ends with'] then
      return strfind(info.name, crit.value..'$') or false
   end
   return false
end

famBags.Classification[famBags.L['Item Level']] = function(crit, info)
   local operator = next(crit.criterias)
   local ilvl = tonumber(crit.value)
   if not (info.iLvl and operator and crit.value and ilvl) then return false end
   if operator == famBags.L['Equal'] then
      return info.iLvl == ilvl
   elseif operator == famBags.L['Greater or equal'] then
      return info.iLvl >= ilvl
   elseif operator == famBags.L['Less or equal'] then
      return info.iLvl <= ilvl
   end
   return false
end

famBags.Classification[famBags.L['Required Level']] = function(crit, info)
   local operator = next(crit.criterias)
   local lvl = tonumber(crit.value)
   if not (info.minLvl and operator and crit.value and lvl) then return false end
   if operator == famBags.L['Equal'] then
      return info.minLvl == lvl
   elseif operator == famBags.L['Greater or equal'] then
      return info.minLvl >= lvl
   elseif operator == famBags.L['Less or equal'] then
      return info.minLvl <= lvl
   end
   return false
end

